#!/usr/bin/tcl
#
# This reads from stdin and writes to stdout.
# First pass removes whitespace, redundant tags,
# and decides on order of tag openings. Second
# pass iterates over the list of records and
# generates output.
#
# I started this on Saturday night, as a rewrite
# of my lightning entry.
#
# Note: After untarring this, you may need to
# chmod a+x the script to run. Also, this
# script requires tcl in /usr/bin/tcl.

#
# Init globals
#

array set defaultState {
    b 0
    em 0
    i 0
    pl 0
    s 0
    tt 0
    u 0
    size ""
    color ""
    text ""
}

set sizeStack {}
set colorStack {}
set plStack {}
set openU 0

set recidx -1

set tagStack {}
array set tagArray {}

set outputBuf ""
set readPos 0

set textBuf ""
set textBufLen 0

set closeList ""
set enabledList ""
set recordList ""
array set startArray {}

###

proc getmem {} {
    set mem 0
    catch {
	set mem [exec /bin/ps -o vsz -p [pid]]
	set mem [lindex $mem 1]
    }
    return $mem
}

proc checkmem {} {
    if {[getmem] > [expr 1000000 * 900]} {
	return 0
    }
    return 1
}

proc scan {} {

    global readPos inputBuf
    global sizeStack colorStack plStack
    global defaultState recidx
    global startTime abort maxtime
    global textBuf textBufLen
    global closeList enabledList recordList startArray

    set readPos 0

    set sizeStack {}
    set colorStack {}
    set plStack {}

    set textBuf ""
    set textBufLen 0

    set closeList ""
    set enabledList ""
    set recordList ""
    array set startArray {}

    set recidx -1
  
    array set state [array get defaultState]
    array set prev [array get defaultState]
    array set roots {b -1 em -1 i -1 pl -1 s -1 tt -1 u -1}

    set wsContext ""
    set prevWsContext ""

    set trailingSpace 0
    set leadingSpace 0
    
    set count 0

    while 1 {

	if {[expr $count % 200] == 0} {
	    if {$maxtime != ""} {
		if {[expr $startTime - [clock seconds]] >= $maxtime} {
		    set abort 1
		    return ""
		}
	    }
	    puts stderr "scan: $count $readPos"
	}
	if {[expr $count % 10000]} {
	    if {![checkmem]} {set abort 1 ; return ""}
	}
	incr count

	#
	# Get text
	#

	set lt [string first "<" $inputBuf $readPos]
	set gt [string first ">" $inputBuf [expr $lt+1]]

	set text ""

	if {$lt > 0} {
	    set text [string range $inputBuf $readPos [expr $lt-1]]
	} elseif {$lt == -1} {
	    set text [string range $inputBuf $readPos end]
	}

	if {$gt >= 0} {set readPos [expr $gt+1]}

	if {$text != ""} {

	    #
	    # Trim whitespace
	    #

	    set leadingSpace 0
	    set newTrailingSpace 0

	    if {!$state(tt)} {
		regsub -all {\s+} $text { } text
	    }
	    if {[regexp {\s} [string index $text 0]]} {
		set leadingSpace 1
	    }
	    if {[regexp {\s} [string index $text end]] && ([string length $text] > 1)} {
		set newTrailingSpace 1
	    }

	    set wsContext ""
	    set wsColor $state(color)
	    if {$state(u) == 0} {set wsColor w}
	    set wsContext [list $state(u) $wsColor $state(size) $state(tt)]

	    if {!$state(tt) && $trailingSpace && $leadingSpace && \
		    ($wsContext == $prevWsContext)} {
		set text [string range $text 1 end]
	    }

	    set prevWsContext $wsContext
	    set trailingSpace $newTrailingSpace

	    #
	    # Save this state
	    #

	    if {$text != ""} {
		set textLen [string length $text]
		append textBuf $text
		set state(text) "$textBufLen [expr ($textBufLen + $textLen)-1]"
		set textBufLen [expr $textBufLen + $textLen]
		incr recidx
		# set state(close) ""
		set closeElem {}
		#set state(start) ""
		# set state(range) ""
		# set state(enabled) ""
		set enabledElem {}
		set startElem {}
		foreach t {b em i s tt} {
		    if {$state($t) != $prev($t)} {
			if {$state($t) > $prev($t)} {
			    set n [expr $state($t) - $prev($t)]
			    for {set j 0} {$j < $n} {incr j} {
				lappend roots($t) $recidx
			    }
			} elseif {$state($t) < $prev($t)} {
			    set started 0
			    set r $roots($t)
			    set n [expr $prev($t) - $state($t)]
			    for {set j 0} {$j < $n} {incr j} {
				if {[llength $roots($t)] <= 0} {break}
				if {$state($t) == 0 && $started == 0} {
				    set r [lindex $roots($t) 0]
				    #global rec_$r
				    #lappend rec_${r}(start) $t
				    # lappend rec_${r}(range) {$t $recidx}
				    #array set rec [lindex $recordList $r]
				    #lappend rec(start) $t
				    #lreplace $recordList $r $r [array get rec]
				    lappend startArray($r) $t
				    lappend closeElem $t
				    set started 1
				}
				set roots($t) [lreplace $roots($t) end end]
			    }
			}
		    }
		    if {$state($t) > 0} {
			set c 1
			lappend enabledElem [list $t $c]
		    }
		}
		lappend closeList $closeElem
		lappend enabledList $enabledElem
		#set rec rec_$recidx ; global $rec
		set st [array get state]
		#array set $rec $st
		lappend recordList $st
		array set prev $st
		set startArray($recidx) ""
	    }
	}

	#
	# Get next tag
	#

	if {$lt == -1 || $gt == -1} {	    
	    break
	}

	set tag [string range $inputBuf $lt $gt]
	set tagLen [string length $tag]
	set tagName ""

	if {[string index $tag 1] == "/"} {
	    set tagName [string range $tag 2 [expr $tagLen-2]]
	    set open 0
	} else {
	    set tagName [string range $tag 1 [expr $tagLen-2]]
	    set open 1
	}

	if {[regexp {^[0-9]$} $tagName]} {
	    if {$open} {
		lappend sizeStack $tagName
		set state(size) $tagName
	    } else {
		set sizeStack [lreplace $sizeStack end end]
		set last [lindex $sizeStack end]
		set state(size) $last
	    }
	} elseif {[regexp {^[rgbcmykw]$} $tagName]} {
	    if {$open} {
		lappend colorStack $tagName
		set state(color) $tagName
	    } else {
		set colorStack [lreplace $colorStack end end]
		set last [lindex $colorStack end]
		set state(color) $last
	    }
	} elseif {$tagName == "PL"} {
	    if {$open} {
		set plList [list u $state(u) b $state(b) em $state(em) \
				i $state(i) s $state(s) tt $state(tt)]
		lappend plStack $plList
		foreach t {u b em i s tt} {set state($t) 0}
	    } else {
		array set plArray [lindex $plStack end]
		if {[array size plArray] > 0} {
		    foreach t {u b em i s tt} {set state($t) $plArray($t)}
		    set plStack [lreplace $plStack end end]
		}
	    }
	} elseif {$tagName == "U"} {
	    set u $state(u)
	    if {$open} {
		set state(u) [expr $u+1]
	    } else {
		set state(u) [expr $u-1]
	    }
	} elseif {$tagName == "EM"} {
	    if {$state(s) == 0} {
		set em $state(em)
		if {$em == 0} {
		    set state(em) 1
		} else {
		    set state(em) 0
		}
	    }
	} elseif {$tagName == "B" || $tagName == "I" || \
		      $tagName == "S" || $tagName == "TT"} {
	    set t [string tolower $tagName]
	    if {$open} {
		set state($t) [expr $state($t)+1]
	    } else {
		set state($t) [expr $state($t)-1]
	    }
	}
    }
}

proc wire {} {

    global defaultState recidx
    global tagStack tagArray outputBuf
    global sizeStack colorStack
    global startTime abort maxtime
    global textBuf textBufLen openU
    global closeList enabledList recordList startArray

    set tagStack {}
    array set tagArray {}
    set outputBuf ""

    set sizeStack {}
    set colorStack {}

    array set state [array get defaultState]
    array set prev [array get defaultState]

    set recCount [expr $recidx+1]
    set recLast $recidx

    set count 0

    #for {set i 0} {$i < $recCount} {incr i} {}

    set i -1

    foreach recl $recordList {

	incr i

	if {[expr $count % 200] == 0} {
	    if {$maxtime != ""} {
		if {[expr $startTime - [clock seconds]] >= $maxtime} {
		    set abort 1
		    return ""
		}
	    }
	    puts stderr "wire: $count"
	}
	if {[expr $count % 10000]} {
	    if {![checkmem]} {set abort 1 ; return ""}
	}
	incr count

	#set rec rec_$i ; global $rec
	#array set state [array get $rec]

	array set state $recl

	set a [lindex $state(text) 0]
	set b [lindex $state(text) 1]
	set text [string range $textBuf $a $b]

	if {$state(size) == "" && $state(size) != $prev(size)} {
	    while 1 {
		set t [lindex $sizeStack 0]
		if {$t == ""} {break}
		tagClose $t
	    }
	}
	if {$state(color) == "" && $state(color) != $prev(color)} {
	    while 1 {
		set t [lindex $colorStack 0]
		if {$t == ""} {break}
		tagClose $t
	    }
	}
	if {$state(u) < $openU} {
	    set n [expr $openU-$state(u)]
	    for {set j 0} {$j<$n} {incr j} {
		tagClose U
	    }
	}
	foreach t [lindex $closeList $i] {
	    tagClose [string toupper $t]
	}

	set state(start) $startArray($i)
	foreach t $state(start) {
	    tagOpen [string toupper $t]
	}
	foreach tv [lindex $enabledList $i] {
	    set t [string toupper [lindex $tv 0]]
	    set v [lindex $tv 1]
	    if {![info exists tagArray($t)]} {
		tagOpen $t
	    }
	    while {$tagArray($t) < $v} {
		tagOpen $t
	    }
	}
	if {$state(u) > $openU && $state(u) <= 3} {
	    set n [expr $state(u)-$openU]
	    for {set j 0} {$j<$n} {incr j} {
		tagOpen U
	    }
	}
	if {$state(color) != "" && $state(color) != [lindex $colorStack end]} {
	    tagOpen $state(color)
	    lappend colorStack $state(color)
	}
	if {$state(size) != "" && $state(size) != [lindex $sizeStack end]} {
	    tagOpen $state(size)
	    lappend sizeStack $state(size)
	}

	append outputBuf $text

	# puts [array get state]

	array set prev [array get state]
    }

    tagFlush
}

proc tagClose {t} {
    global tagStack tagArray outputBuf
    global sizeStack colorStack openU
    set i 1
    while 1 {
	set pop [lindex $tagStack end]
	if {$pop == ""} {break}
	set tagStack [lreplace $tagStack end end]
	if {[info exists tagArray($pop)] && ($tagArray($pop) > 0)} {
	    set tagArray($pop) [expr $tagArray($pop)-1]
	} else {
	    set tagArray($pop) 0
	}
	append outputBuf "</$pop>"
	if {[regexp {^[0-9]$} $pop]} {
	    set sizeStack [lreplace $sizeStack end end]
	}
	if {[regexp {^[rgbcmykw]$} $pop]} {
	    set colorStack [lreplace $colorStack end end]
	}
	if {$pop == "U"} {set openU [expr $openU-1]}
	if {$pop == $t} {break}
    }
}

proc tagOpen {t} {
    global tagStack tagArray outputBuf openU
    lappend tagStack $t
    if {[info exists tagArray($t)]} {
	set tagArray($t) [expr $tagArray($t)+1]
    } else {
	set tagArray($t) 1
    }
    if {$t == "U"} {incr openU}
    append outputBuf "<$t>"
}

proc tagFlush {} {
    global tagStack outputBuf
    while 1 {
	set last [lindex $tagStack end]
	if {$last == ""} {break}
	tagClose $last
    }
}

proc abort {} {
    global inputBuf
    puts -nonewline $inputBuf
    exit 0
}

#
# Begin
#

set startTime [clock seconds]

#
# Get timeout
#

set maxtime ""
set abort 0
set arg1 [lindex $argv 0]
if {[string is integer $arg1] && $arg1 != ""} {
    set maxtime $arg1
}
if {[info exists env(TIMEOUT)]} {
    set timeout $env(TIMEOUT)
    if {[string is integer $timeout] && $timeout != ""} {
	set maxtime $timeout
    }
}
if {$maxtime != ""} {
    set minus [expr ceil($maxtime.0 / 10.0)]
    if {$minus < 3} {set minus 3}
    if {$minus > 10} {set minus 10}
    set maxtime [expr $maxtime - $minus]
    if {$maxtime <= 0} {
	set maxtime 0
	set abort 1
    }
}

#
# Read file
#

puts -nonewline stderr "reading:"
if {[lindex $argv 0] != "."} {
    set inputBuf [read stdin]
}
set inputLen [string length $inputBuf]
puts stderr " $inputLen"

if {$abort == 1} {abort}

#
# Process input
#

scan
if {$abort == 1} {abort}
wire
if {$abort == 1} {abort}

#
# Send output
#

set outputLen [string length $outputBuf]

if {$outputLen > $inputLen} {
    puts stderr ">>> $inputLen $outputLen"
    set outputBuf $inputBuf
    set outputLen $inputLen
}

puts -nonewline $outputBuf

set stopTime [clock seconds]
set processTime [expr $stopTime - $startTime]

puts stderr "$inputLen $outputLen $processTime"
